#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <algorithm>
#include <cassert>
#include <cmath>
#include <format>
#include <ostream>
#include <J3ML/Geometry/AABB.hpp>
#include <J3ML/Geometry/Sphere.hpp>

namespace J3ML::LinearAlgebra {

    const Vector3 Vector3::Zero     = {0,0,0};
    const Vector3 Vector3::Up       = {0, -1, 0};
    const Vector3 Vector3::Down     = {0, 1, 0};
    const Vector3 Vector3::Left     = {-1, 0, 0};
    const Vector3 Vector3::Right    = {1, 0, 0};
    const Vector3 Vector3::Forward  = {0, 0, -1};
    const Vector3 Vector3::Backward = {0, 0, 1};
    const Vector3 Vector3::NaN      = {NAN, NAN, NAN};
    const Vector3 Vector3::Infinity = {INFINITY, INFINITY, INFINITY};
    const Vector3 Vector3::NegativeInfinity = {-INFINITY, -INFINITY, -INFINITY};
    const Vector3 Vector3::UnitX = {1,0,0};
    const Vector3 Vector3::UnitY = {0,1,0};
    const Vector3 Vector3::UnitZ = {0,0,1};

    Vector3 Vector3::operator+(const Vector3& rhs) const
    {
        return {this->x + rhs.x, this->y + rhs.y, this->z + rhs.z};
    }

    Vector3 Vector3::operator-(const Vector3& rhs) const
    {
        return {
                this->x- rhs.x,
                this->y-rhs.y,
                this->z-rhs.z
        };
    }

    Vector3 Vector3::operator*(float rhs) const
    {
        return {
                this->x * rhs,
                this->y * rhs,
                this->z * rhs
        };
    }


    Vector3 Vector3::operator/(float rhs) const
    {
        return {
                this->x / rhs,
                this->y / rhs,
                this->z / rhs
        };
    }

    Vector3 Vector3::operator-() const
    {
        return {-x, -y, -z};
    }




    Vector3::Vector3(): x(0), y(0), z(0) {}

    Vector3::Vector3(float X, float Y, float Z): x(X), y(Y), z(Z) {}

    //Vector3::Vector3(const Vector3& rhs) : x(rhs.x), y(rhs.y), z(rhs.z) {}

    Vector3& Vector3::operator=(const Vector3& rhs)
    {
        this->x = rhs.x;
        this->y = rhs.y;
        this->z = rhs.z;
        return *this;
    }

    float Vector3::operator[](std::size_t index) const
    {
        assert(index < 3);
        if (index==0) return x;
        if (index==1) return y;
        if (index==2) return z;
        return 0;
    }

    float &Vector3::operator[](std::size_t index)
    {
        assert(index < 3);
        if (index == 0) return x;
        if (index == 1) return y;
        if (index == 2) return z;
        throw;
    }

    bool Vector3::IsWithinMarginOfError(const Vector3& rhs, float margin) const
    {
        return this->Distance(rhs) <= margin;
    }

    bool Vector3::operator==(const Vector3& rhs) const
    {
        return this->IsWithinMarginOfError(rhs);
    }

    bool Vector3::operator!=(const Vector3& rhs) const
    {
        return !this->IsWithinMarginOfError(rhs);
    }



    Vector3 Vector3::Min(const Vector3& min) const
    {
        return {
                std::min(this->x, min.x),
                std::min(this->y, min.y),
                std::min(this->z, min.z)
        };
    }

    Vector3 Vector3::Max(const Vector3& max) const
    {
        return {
                std::max(this->x, max.x),
                std::max(this->y, max.y),
                std::max(this->z, max.z)
        };
    }

    Vector3 Vector3::Clamp(const Vector3& min, const Vector3& max) const
    {
        return {
                std::clamp(this->x, min.x, max.x),
                std::clamp(this->y, min.y, max.y),
                std::clamp(this->z, min.z, max.z)
        };
    }

    float Vector3::Distance(const Vector3& to) const
    {
        return ((*this)-to).Magnitude();
    }

    float Vector3::Length() const
    {
        return std::sqrt(LengthSquared());
    }

    float Vector3::LengthSquared() const
    {
        return (x*x + y*y + z*z);
    }

    float Vector3::Magnitude() const
    {
        return std::sqrt(x*x + y*y + z*z);
    }

    float Vector3::Dot(const Vector3& rhs) const
    {
        return x * rhs.x +
               y * rhs.y +
               z * rhs.z;
    }

    Vector3 Vector3::Project(const Vector3& rhs) const
    {
        float scalar = this->Dot(rhs) / (rhs.Magnitude()*rhs.Magnitude());
        return rhs * scalar;
    }

    Vector3 Vector3::Cross(const Vector3& rhs) const
    {
        return {
                this->y * rhs.z - this->z * rhs.y,
                this->z * rhs.x - this->x * rhs.z,
                this->x * rhs.y - this->y * rhs.x
        };
    }

    float Vector3::Normalize()
    {
        assert(IsFinite());
        float length = Length();
        if (length > 1e-6f)
        {
            *this *= 1.f / length;
            return length;
        } else {
            Set(1.f, 0.f, 0.f); // We will always produce a normalized vector.
            return 0; // But signal failure, so user knows we have generated an arbitrary normalization.
        }
    }

    Vector3 Vector3::Normalized() const
    {
        if (Length() > 0)
            return {
                    x / Length(),
                    y / Length(),
                    z / Length()
            };
        else
            return {0,0,0};
    }

    Vector3 Vector3::Lerp(const Vector3& goal, float alpha) const
    {
        return this->operator*(1.0f - alpha) + (goal * alpha);
    }

    float Vector3::GetX() const { return x;}

    float Vector3::GetY() const { return y;}

    float Vector3::GetZ() const { return z;}

    void Vector3::SetX(float newX) { x = newX;}

    void Vector3::SetY(float newY) { y = newY;}

    void Vector3::SetZ(float newZ) { z = newZ;}

    Vector3 Vector3::Min(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Min(rhs);
    }

    Vector3 Vector3::Max(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Max(rhs);
    }

    Vector3 Vector3::Clamp(const Vector3 &min, const Vector3 &input, const Vector3 &max) {
        return input.Clamp(min, max);
    }

    float Vector3::Distance(const Vector3 &from, const Vector3 &to) {
        return from.Distance(to);
    }

    float Vector3::Length(const Vector3 &of) {
        return of.Length();
    }

    float Vector3::LengthSquared(const Vector3 &of) {
        return of.LengthSquared();
    }

    bool Vector3::IsPerpendicular(const Vector3 &other, float epsilonSq) const {
        float dot = Dot(other);
        return dot*dot <= epsilonSq * LengthSquared() * other.LengthSquared();
    }

    bool Vector3::IsZero(float epsilonSq) const {
        return LengthSquared() <= epsilonSq;
    }

    bool Vector3::IsNormalized(float epsilonSq) const {
        return std::abs(LengthSquared()-1.f) <= epsilonSq;
    }

    Vector3 Vector3::Cross(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Cross(rhs);
    }

    Vector3 Vector3::Normalized(const Vector3 &targ) {
        return targ.Normalized();
    }

    Vector3 Vector3::Project(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Project(rhs);
    }

    float Vector3::Dot(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Dot(rhs);
    }

    float Vector3::Magnitude(const Vector3 &of) {
        return of.Magnitude();
    }

    Vector3 Vector3::Lerp(const Vector3 &lhs, const Vector3 &rhs, float alpha) {
        return lhs.Lerp(rhs, alpha);
    }

    Vector3 Vector3::Add(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Add(rhs);
    }

    Vector3 Vector3::Add(const Vector3 &rhs) const {
        return *this + rhs;
    }

    Vector3 Vector3::Sub(const Vector3 &rhs) const {
        return *this - rhs;
    }

    Vector3 Vector3::Sub(const Vector3 &lhs, const Vector3 &rhs) {
        return lhs.Sub(rhs);
    }

    Vector3 Vector3::Mul(float scalar) const {
        return *this * scalar;
    }

    Vector3 Vector3::Mul(const Vector3 &lhs, float rhs) {
        return lhs.Mul(rhs);
    }

    Vector3 Vector3::Div(float scalar) const {
        return *this / scalar;
    }

    Vector3 Vector3::Div(const Vector3 &lhs, float rhs) {
        return lhs.Div(rhs);
    }

    Angle2D Vector3::AngleBetween(const Vector3 &rhs) const {
        const auto Pi_x_180 = 180.f / Math::Pi;
        auto dist = this->Distance(rhs);
        float dx = -(asinf((rhs.y - this->y) / dist));
        float dy = (atan2f(rhs.x - this->x,rhs.z - this->z));
        return {dx, dy};
    }

    Angle2D Vector3::AngleBetween(const Vector3 &lhs, const Vector3 &rhs) // TODO: 3D Angle representation?
    {
        return lhs.AngleBetween(rhs);
    }

    Vector3 Vector3::Direction(const Vector3 &rhs) {
        float x = (Math::Cos(Math::Radians(rhs.y)) * Math::Cos(Math::Radians(rhs.x)));
        float y = -Math::Sin(Math::Radians(rhs.x));
        float z = (Math::Sin(Math::Radians(rhs.y)) * Math::Cos(Math::Radians(rhs.x)));
        return {x, y, z};
    }

    float Vector3::ScaleToLength(float newLength) {
        float length = LengthSquared();
        if (length < 1e-6f) {
            Set(newLength, 0, 0); // Will always produce a vector of the requested length.
            return 0.f;
        }

        length = Math::Sqrt(length);
        float scalar = newLength / length;
        x *= scalar;
        y *= scalar;
        z *= scalar;

        return length;
    }

    Vector3 Vector3::Abs() const {
        return {std::abs(x), std::abs(y), std::abs(z)};
    }

    float *Vector3::ptr() {
        return &x;
    }

    void Vector3::Orthonormalize(Vector3 &a, Vector3 &b) {
        a = a.Normalized();
        b = b - b.ProjectToNorm(a);
        b = b.Normalized();
    }

    void Vector3::Orthonormalize(Vector3 &a, Vector3 &b, Vector3 &c) {
        a = a.Normalized();
        b = b - b.ProjectToNorm(a);
        b = b.Normalized();
        c = c - c.ProjectToNorm(a);
        c = c - c.ProjectToNorm(b);
        c = c.Normalized();
    }

    Vector3 Vector3::ProjectToNorm(const Vector3 &direction) const {
        return direction * this->Dot(direction);
    }

    bool Vector3::IsFinite() const {
        return std::isfinite(x) && std::isfinite(y) && std::isfinite(z);
    }

    Vector3::Vector3(const float *data) {
        x = data[0];
        y = data[1];
        z = data[2];
    }

    Vector3 &Vector3::operator+=(const Vector3 &rhs) {
        x += rhs.x;
        y += rhs.y;
        z += rhs.z;
        return *this;
    }

    Vector3 &Vector3::operator-=(const Vector3 &rhs) {
        x -= rhs.x;
        y -= rhs.y;
        z -= rhs.z;
        return *this;
    }

    Vector3 &Vector3::operator*=(float scalar) {
        x *= scalar;
        y *= scalar;
        z *= scalar;
        return *this;
    }

    Vector3 &Vector3::operator/=(float scalar) {
        x /= scalar;
        y /= scalar;
        z /= scalar;
        return *this;
    }

    float Vector3::MinElement() const {
        return std::min(x, std::min(y, z));
    }

    bool Vector3::AreOrthonormal(const Vector3 &a, const Vector3 &b, float epsilon) {
        return a.IsPerpendicular(b, epsilon) && a.IsNormalized(epsilon*epsilon) && b.IsNormalized(epsilon*epsilon);
    }

    Vector3 Vector3::Mul(const Vector3 &rhs) const {
        return {
                this->x * rhs.x,
                this->y * rhs.y,
                this->z * rhs.z
        };
    }

    Vector3 Vector3::Div(const Vector3 &v) const {
        return {
                this->x/v.x,
                this->y/v.y,
                this->z/v.z
        };
    }

    Vector3 Vector3::Abs(const Vector3 &rhs) {
        return rhs.Abs();
    }

    bool Vector3::Equals(const Vector3 &rhs, float epsilon) const {
        return std::abs(x - rhs.x) < epsilon &&
               std::abs(y - rhs.y) < epsilon &&
               std::abs(z - rhs.z) < epsilon;
    }

    bool Vector3::Equals(float _x, float _y, float _z, float epsilon) const {
        return std::abs(x - _x) < epsilon &&
               std::abs(y - _y) < epsilon &&
               std::abs(z - _z) < epsilon;
    }

    Vector3 Vector3::Min(const Vector3 &a, const Vector3 &b, const Vector3 &c) {
        return {
                std::min(a.x, std::min(b.x, c.x)),
                std::min(a.y, std::min(b.y, c.y)),
                std::min(a.z, std::min(b.z, c.z))
        };
    }

    Vector3 Vector3::Max(const Vector3 &a, const Vector3 &b, const Vector3 &c) {
        return {
                std::max(a.x, std::max(b.x, c.x)),
                std::max(a.y, std::max(b.y, c.y)),
                std::max(a.z, std::max(b.z, c.z))
        };
    }

    // TODO: Implement
    Vector3 Vector3::Perpendicular(const Vector3 &hint, const Vector3 &hint2) const {
        assert(!this->IsZero());
        assert(hint.IsNormalized());
        assert(hint2.IsNormalized());
        Vector3 v = this->Cross(hint);
        float len = v.TryNormalize();


        return Vector3::Zero;
    }


    Vector3 Vector3::AnotherPerpendicular(const J3ML::LinearAlgebra::Vector3 &hint,
                                          const J3ML::LinearAlgebra::Vector3 &hint2) const {
        assert(!this->IsZero());
        assert(hint.IsNormalized());
        assert(hint2.IsNormalized());
        Vector3 firstPerpendicular = Perpendicular(hint, hint2);
        Vector3 v = this->Cross(firstPerpendicular);
        return v.Normalized();
    }

    float Vector3::TryNormalize() {
        assert(IsFinite());
        float length = Length();
        if (length > 1e-6f)
        {
            *this *= 1.f / length;
            return length;
        }
        else
        {
            Set(1.f, 0.f, 0.f); // We will always produce a normalized vector.
            return 0; // But signal failure, so user knows we have generated an arbitrary normalization.
        }
    }

    float Vector3::At(int index) const {
        assert(index >= 0);
        assert(index < Dimensions);
        return ptr()[index];
    }

    float &Vector3::At(int index) {
        assert(index >= 0);
        assert(index < Dimensions);
        return ptr()[index];
    }

    void Vector3::Set(float d, float d1, float d2) {
        x = d;
        y = d1;
        z = d2;
    }

    std::string Vector3::ToString() const {
        return std::format("{},{},{}", x, y, z);
    }

    std::ostream& operator<<(std::ostream &o, const Vector3 &vector) {
        o << vector.ToString();
        return o;
    }

    Vector3::Vector3(const Vector2& XY, float Z) {
        x = XY.x;
        y = XY.y;
        z = Z;
    }

    Vector3::Vector3(float scalar) {
        x = scalar;
        y = scalar;
        z = scalar;
    }

    float Vector3::DistanceSquared(const Vector3 &to) const {
        return (*this-to).LengthSquared();
    }

    float Vector3::DistanceSquared(const Vector3 &from, const Vector3 &to) {
        return from.DistanceSquared(to);
    }

    bool Vector3::AreOrthonormal(const Vector3 &a, const Vector3 &b, const Vector3 &c, float epsilon) {
        return a.IsPerpendicular(b, epsilon) &&
               a.IsPerpendicular(c, epsilon) &&
               b.IsPerpendicular(c, epsilon) &&
               a.IsNormalized(epsilon*epsilon) &&
               b.IsNormalized(epsilon*epsilon) &&
               c.IsNormalized(epsilon*epsilon);
    }

    Vector3 Vector3::FromScalar(float scalar) {
        return {scalar, scalar, scalar};
    }

    Vector3 Vector3::RandomDir(RNG &rng, float length) {
        return Sphere(Vector3::FromScalar(0.f), length).RandomPointOnSurface(rng);
    }

    Vector3 Vector3::RandomSphere(RNG &rng, const Vector3 &center, float radius) {
        return Sphere(center, radius).RandomPointInside(rng);
    }

    Vector3 Vector3::RandomBox(RNG &rng, const Vector3& min, const Vector3& max) {
        return AABB(min, max).RandomPointInside(rng);
    }

    Vector3 Vector3::RandomBox(RNG &rng, float min, float max) {
        return RandomBox(rng, {min,min,min}, {max,max,max});
    }

    Vector3 Vector3::RotateAroundAxis(const Vector3 &vec, const Vector3 &axis, const float radians) {
        float cosTheta = cos(radians);
        float sinTheta = sin(radians);
        return vec * cosTheta + Vector3::Cross(axis, vec) * sinTheta + axis * Vector3::Dot(axis, vec) * (1 - cosTheta);
    }

    Vector3 Vector3::RandomBox(RNG &rng, float xmin, float xmax, float ymin, float ymax, float zmin, float zmax) {
        return RandomBox(rng, {xmin,ymin,zmin}, {xmax,ymax,zmax});
    }

    Vector3 Vector3::Add(float s) const {
        return *this + Vector3(s, s, s);
    }

    Vector3 Vector3::ScaledToLength(float newLength) const {
        assert(!IsZero());

        Vector3 v = *this;
        v.ScaleToLength(newLength);
        return v;
    }

    void Vector3::PerpendicularBasis(Vector3 &outB, Vector3 &outC) const {
        // Pixar orthonormal basis code: https://graphics.pixar.com/library/OrthonormalB/paper.pdf
        float sign = copysignf(1.0f, z);
        const float a = -1.0f / (sign + z);
        const float b = x * y * a;
        outB = Vector3(1.0f + sign * x * x * a, sign * b,             -sign * x);
        outC = Vector3(                      b, sign + y * y * a,            -y);
    }


}